import qlat_utils.c as c

from qlat_utils.cache import *
from qlat_utils.rng_state import *

import math
import sys
import os

def getenv(*names, default = None):
    assert len(names) > 0
    for name in names:
        val = os.getenv(name)
        if val is not None:
            displayln_info(0, f"{name}='{val}'")
            return val
    val = default
    displayln_info(0, f"{names[0]}='{val}' (default)")
    return val

def get_arg(option, default = None):
    argv = sys.argv
    i_max = len(argv) - 1
    for i, arg in enumerate(argv):
        if arg == option:
            if i == i_max:
                return ""
            else:
                return argv[i + 1]
    return default

def random_permute(l, rs):
    # Do not change ``l''.
    # Return a new permuted list.
    assert isinstance(l, list)
    assert isinstance(rs, RngState)
    return c.random_permute(l, rs)

def show_memory_usage():
    try:
        import psutil
        rss = psutil.Process().memory_info().rss / (1024 * 1024 * 1024)
        displayln_info(f"show_memory_usage: rss = {rss:.6f} GB")
        # displayln_info_malloc_stats()
    except:
        displayln_info(f"show_memory_usage: no psutil.")

def displayln_malloc_stats():
    return c.displayln_malloc_stats()

def displayln_info_malloc_stats():
    if get_id_node() == 0:
        return displayln_malloc_stats()

def lazy_call(f, *args, **kwargs):
    is_thunk = True
    ret = None
    def get():
        nonlocal ret, is_thunk
        if is_thunk:
            ret = f(*args, **kwargs)
            is_thunk = False
        return ret
    return get

def sqr(x):
    return x * x

def set_zero(x):
    x.set_zero()

def set_unit(x, coef = 1.0):
    x.set_unit(coef)

def show(x):
    return x.show()

def unitarize(x):
    x.unitarize()

def rel_mod(x, size):
    x = x % size
    assert x >= 0
    if 2 * x >= size:
        return x - size
    else:
        return x

def rel_mod_sym(x, size):
    x = x % size
    assert x >= 0
    if 2 * x > size:
        return x - size
    elif 2 * x < size:
        return x
    else:
        assert 2 * x == size
        return 0

def c_sqr(x):
    return sum([ sqr(v) for v in x ])

def c_rel_mod(x, size):
    l = len(size)
    assert l == len(x)
    return [ rel_mod(x[i], size[i]) for i in range(l) ]

def c_rel_mod_sqr(x, size):
    l = len(size)
    assert l == len(x)
    return sum([ sqr(rel_mod(x[i], size[i])) for i in range(l) ])

def phat_sqr(q, size):
    l = len(size)
    assert l == len(q)
    return 4 * sum([ sqr(math.sin(math.pi * (q[i] % size[i]) / size[i])) for i in range(l) ])



from libcpp.string cimport string as std_string
from libcpp cimport bool

cdef extern from "qlat-utils/timer.h" namespace "qlat":

    int get_id_node()
    int get_num_node()

    long& verbose_level()

    double get_time()
    double& get_start_time()
    double& get_actual_start_time()

    double get_total_time()
    double get_actual_total_time()

    cdef cppclass Timer:
        long long flops
        Timer()
        Timer(const std_string& fname)
        void start()
        void start(bool is_verbose)
        void stop()
        void stop(bool is_verbose)
        @staticmethod
        void display(const std_string& tag)
        @staticmethod
        void autodisplay()
        @staticmethod
        void display_stack()
        @staticmethod
        void display_stack_always()
        @staticmethod
        void reset(long max_call_times_for_always_show_info)
        @staticmethod
        void fork(long max_call_times_for_always_show_info)
        @staticmethod
        void merge()



from .complex cimport *
from .vector cimport *
from .timer cimport *
from .utils_io cimport *
from .rng_state cimport *
from .coordinate cimport *
from .elem_type cimport *
from .lat_io cimport *
from .mat cimport *



from .rng_state cimport *

cdef extern from "qlat-utils/coordinate.h" namespace "qlat":

    cdef cppclass Coordinate:
        Coordinate()
        Coordinate(int x, int y, int z, int t)
        int& operator[](unsigned long i)

    Coordinate coordinate_from_index(long index, const Coordinate& size)
    long index_from_coordinate(const Coordinate& x, const Coordinate& size)
    int eo_from_coordinate(const Coordinate& xl)

    Coordinate mod(const Coordinate& x, const Coordinate& size)
    Coordinate smod(const Coordinate& x, const Coordinate& size)
    Coordinate middle_mod(const Coordinate& x, const Coordinate& size)

    Coordinate c_rand_gen(RngState& rs, const Coordinate& size)

    cdef cppclass CoordinateD:
        CoordinateD()
        CoordinateD(const Coordinate& x)
        CoordinateD(double x, double y, double z, double t)
        double& operator[](unsigned long i)



from qlat_utils.timer import *

import qlat_utils.c as c

# Usage:
# cache_x = q.mk_cache("xx")
# q.clean_cache(cache_x)
# cache_x[key] = value
# val = cache_x[key]
# key in cache_x
# val = cache_x.get(key)
# val = cache_x.pop(key, None)

class Cache(dict):

    def __init__(self, *keys):
        super().__init__()
        self.keys = keys

###

cache = Cache()

@timer
def list_cache(ca = cache):
    l = dict()
    for key, val in ca.items():
        if isinstance(val, Cache):
            l[key] = list_cache(val)
        else:
            l[key] = None
    return l

def show_cache_keys(keys):
    if not keys:
        return ""
    else:
        return "['" + "']['".join(keys) + "']"

@timer
def clean_cache(ca = cache):
    info_str = show_cache_keys(ca.keys)
    items = list(ca.items())
    displayln_info(0, f"clean_cache: cache{info_str}: len={len(items)}")
    for key, val in items:
        if isinstance(val, Cache):
            clean_cache(val)
        else:
            # displayln_info(1, f"clean_cache: cache{info_str}['{key}']")
            ca.pop(key)

def mk_cache(*keys, ca = cache):
    # if not exist, otherwise return existing elements
    assert keys
    for key in keys:
        if key in ca:
            ca = ca[key]
            assert isinstance(ca, Cache)
        else:
            ca[key] = Cache(*(ca.keys), key)
            ca = ca[key]
    return ca

@timer
def rm_cache(*keys, ca = cache):
    # if exist
    assert keys
    for key in keys[-1]:
        if key not in ca:
            return
        ca = ca[key]
        assert isinstance(ca, Cache)
    key = keys[-1]
    if key not in ca:
        return
    assert isinstance(ca[key], Cache)
    ca.pop(key)

def get_all_caches_info():
    return c.get_all_caches_info()

def clear_all_caches():
    # clean python level cache and then C++ level cache
    clean_cache()
    c.clear_all_caches()



from qlat_utils.complex cimport *
from qlat_utils.mat cimport *

cdef fused ElemType:
    ColorMatrix
    WilsonMatrix 
    NonRelWilsonMatrix
    SpinMatrix
    WilsonVector
    Complex
    ComplexF
    Double
    Float
    Long
    Int64t
    Int8t
    Char



import qlat_utils.c as c

from qlat_utils.c import flush

from qlat_utils.c import get_id_node, get_num_node
from qlat_utils.c import verbose_level
from qlat_utils.c import get_time, get_start_time, get_actual_start_time
from qlat_utils.c import get_total_time, get_actual_total_time

from qlat_utils.c import timer_display, timer_autodisplay
from qlat_utils.c import timer_display_stack, timer_display_stack_always
from qlat_utils.c import timer_reset, timer_fork, timer_merge

from qlat_utils.c import timer, timer_verbose
from qlat_utils.c import timer_flops, timer_verbose_flops

from qlat_utils.c import Timer, TimerNone

from qlat_utils.c import displayln, displayln_info



from qlat_utils.lat_io import *
from qlat_utils.rng_state import *
from qlat_utils.timer import *

import math
import copy
import functools
import numpy as np

alpha_qed = 1.0 / 137.035999084
fminv_gev = 0.197326979 # hbar * c / (1e-15 m * 1e9 electron charge * 1 volt)

class use_kwargs:

    # self.default_kwargs
    # self.keys

    def __init__(self, default_kwargs, keys = None):
        self.default_kwargs = default_kwargs
        self.keys = None

    def __call__(self, func):
        @functools.wraps(func)
        def f(*args, **kwargs):
            if "is_default_kwargs_applied" not in kwargs:
                kwargs = self.default_kwargs | kwargs
            if self.keys is not None:
                kwargs = { k: kwargs[k] for k in self.keys }
            return func(*args, **kwargs)
        return f

###

def interpolate_list(v, i):
    size = len(v)
    i1 = math.floor(i)
    assert i1 >= 0
    i2 = i1 + 1
    if i2 >= size:
        return v[size - 1]
    elif i1 < 0:
        return v[0]
    v1 = v[i1]
    v2 = v[i2]
    a1 = i2 - i
    a2 = i - i1
    return a1 * v1 + a2 * v2

def interpolate(v_arr, i_arr):
    vt = v_arr.transpose()
    if isinstance(i_arr, (int, float,)):
        return interpolate_list(vt, i_arr).transpose()
    else:
        return np.array([ interpolate_list(vt, i) for i in i_arr ]).transpose()

def partial_sum_list(x, *, is_half_last = False):
    """Modify in-place, preserve length"""
    s = 0
    for i, v in enumerate(x):
        sp = s
        s += v
        if is_half_last:
            x[i] = (s + sp) / 2
        else:
            x[i] = s

def partial_sum(x, *, is_half_last = False):
    """Modify in-place, preserve length"""
    shape = x.shape
    if len(shape) == 0:
        return
    elif len(shape) == 1:
        partial_sum_list(x, is_half_last = is_half_last)
    elif len(shape) == 2:
        for v in x:
            partial_sum_list(v, is_half_last = is_half_last)
    else:
        assert False

def check_zero(x):
    if isinstance(x, (int, float, complex)) and 0 == x:
        return True
    return False

def qnorm(x):
    # qnorm(2) == 4
    if isinstance(x, np.ndarray):
        return np.abs(np.vdot(x, x))
    elif isinstance(x, (int, float,)):
        return x * x
    elif isinstance(x, complex):
        return x.real * x.real + x.imag * x.imag
    elif isinstance(x, (list, tuple,)):
        return sum([ qnorm(x_i) for x_i in x ])
    else:
        return x.qnorm()
    assert False

class Data:

    def __init__(self, val):
        # supported value types:
        # numeric
        # numpy.array
        # q.LatData
        # list
        if isinstance(val, Data):
            self.val = val.val
            assert not isinstance(self.val, Data)
        else:
            self.val = val

    def __str__(self):
        return f"Data({self.val})"

    def get_val(self):
        return self.val

    def __copy__(self):
        return Data(copy.copy(self.val))

    def __deepcopy__(self, memo):
        return Data(copy.deepcopy(self.val, memo))

    def __add__(self, other):
        if isinstance(other, Data):
            if check_zero(self.val):
                return other
            elif check_zero(other.val):
                return self
            elif isinstance(self.val, list) and isinstance(other.val, list):
                assert len(self.val) == len(other.val)
                return Data([ v1 + v2 for v1, v2 in zip(self.val, other.val) ])
            elif isinstance(self.val, list):
                return Data([ v + other.val for v in self.val ])
            elif isinstance(other.val, list):
                return Data([ self.val + v for v in other.val ])
            else:
                return Data(self.val + other.val)
        else:
            return self + Data(other)

    def __radd__(self, other):
        if isinstance(other, Data):
            assert False
            return None
        else:
            return Data(other) + self

    def __mul__(self, other):
        if isinstance(other, Data):
            if check_zero(self.val) or check_zero(other.val):
                return Data(0)
            elif isinstance(self.val, list) and isinstance(other.val, list):
                return Data([ v1 * v2 for v1, v2 in zip(self.val, other.val) ])
            elif isinstance(self.val, list):
                return Data([ v * other.val for v in self.val ])
            elif isinstance(other.val, list):
                return Data([ self.val * v for v in other.val ])
            return Data(self.val * other.val)
        else:
            return self * Data(other)

    def __rmul__(self, other):
        if isinstance(other, Data):
            assert False
            return None
        else:
            return Data(other) * self

    def __neg__(self):
        if check_zero(self.val):
            return Data(0)
        elif isinstance(self.val, list):
            return Data([ -v for v in self.val ])
        else:
            return Data(-self.val)

    def __pos__(self):
        return self

    def __sub__(self, other):
        if isinstance(other, Data):
            if check_zero(self.val):
                return Data(-other.val)
            elif check_zero(other.val):
                return self
            elif isinstance(self.val, list) and isinstance(other.val, list):
                return Data([ v1 - v2 for v1, v2 in zip(self.val, other.val) ])
            elif isinstance(self.val, list):
                return Data([ v - other.val for v in self.val ])
            elif isinstance(other.val, list):
                return Data([ self.val - v for v in other.val ])
            else:
                return Data(self.val - other.val)
        else:
            return self - Data(other)

    def __rsub__(self, other):
        if isinstance(other, Data):
            assert False
            return None
        else:
            return Data(other) - self

    def qnorm(self):
        return qnorm(self.val)

    def glb_sum(self):
        from qlat.mpi import glb_sum
        return Data(glb_sum(self.val))

###

def add_jk_idx(arr):
    # arr: no jk index
    # return: add trivial jk index in the LAST axis
    return arr.reshape(arr.shape + (1,))

def jk_transpose(arr):
    # arr: jk index is the 0th axis
    # return: jk index is the last axis
    shape = arr.shape
    ndim = len(shape)
    if ndim <= 1:
        return arr
    axes = list(range(1, ndim)) + [ 0, ]
    return arr.transpose(axes)

def jk_transpose_back(arr):
    # jk_transpose_back(jk_transpose(arr)) == arr
    shape = arr.shape
    ndim = len(shape)
    if ndim <= 1:
        return arr
    axes = [ ndim - 1, ] + list(range(0, ndim - 1))
    return arr.transpose(axes)

def average(data_list):
    n = len(data_list)
    v = sum(data_list)
    return 1/n * v

def block_data(data_list, block_size, is_overlapping = True):
    # return the list of block averages
    # the blocks may overlap if is_overlapping == True
    if block_size == 1:
        return data_list
    assert block_size >= 1
    size = len(data_list)
    if block_size >= size:
        return [ average(data_list), ]
    blocks = []
    start = 0
    stop = block_size
    while stop <= size:
        b = average(data_list[start:stop])
        blocks.append(b)
        if is_overlapping:
            start += 1
            stop += 1
        else:
            start += block_size
            stop += block_size
    return blocks

def avg_err(data_list, eps = 1, *, block_size = 1):
    avg = average(data_list)
    blocks = block_data(data_list, block_size)
    diff_sqr = average([ fsqr(d - avg) for d in blocks ])
    fac = eps * math.sqrt(block_size / (len(data_list) - 1))
    err = fac * fsqrt(diff_sqr)
    return (avg, err,)

def jackknife(data_list, eps = 1):
    """Return jk[i] = avg + \\frac{eps}{N} \\sum_{i=1}^N (v[i] - avg)
    normal jackknife uses eps = 1, scale the fluctuation by eps"""
    data_list_real = [ d for d in data_list if d is not None ]
    n = len(data_list_real)
    fac = eps / n
    avg = average(data_list_real)
    jks = [ avg, ]
    for data in data_list:
        if data is None:
            jks.append(avg)
        else:
            jks.append(avg + fac * (data - avg))
    return jks

def fsqr(data):
    if isinstance(data, (float, np.float128, int,)):
        return data * data
    elif isinstance(data, (complex, np.complex256,)):
        r = data.real
        i = data.imag
        return complex(r * r, i * i)
    elif isinstance(data, np.ndarray):
        if data.dtype in [ np.float64, np.float128, np.int64, ]:
            return np.square(data)
        elif data.dtype in [ np.complex128, np.complex256, ]:
            return np.square(data.real) + 1j * np.square(data.imag)
        else:
            raise Exception(f"fsqr data={data} type not supported")
    elif isinstance(data, Data):
        return Data(fsqr(data.val))
    else:
        raise Exception(f"fsqr data={data} type not supported")

def fsqrt(data):
    if isinstance(data, (float, np.float128, int,)):
        return math.sqrt(data)
    elif isinstance(data, (complex, np.complex256,)):
        r = data.real
        i = data.imag
        return complex(math.sqrt(r), math.sqrt(i))
    elif isinstance(data, np.ndarray):
        if data.dtype in [ np.float64, np.float128, np.int64, ]:
            return np.sqrt(data)
        elif data.dtype in [ np.complex128, np.complex256, ]:
            return np.sqrt(data.real) + 1j * np.sqrt(data.imag)
        else:
            raise Exception(f"fsqr data={data} type not supported")
    elif isinstance(data, Data):
        return Data(fsqrt(data.val))
    else:
        raise Exception(f"fsqr {data} type not supported")

def jk_avg(jk_list):
    return jk_list[0]

def jk_err(jk_list, eps = 1, *, block_size = 1):
    """Return \\frac{1}{eps} \\sqrt{ \\sum_{i=1}^N (jk[i] - jk_avg)^2 } when block_size=1
    Note: len(jk_list) = N + 1
    Same eps as the eps used in the 'jackknife' function"""
    avg = jk_avg(jk_list)
    blocks = block_data(jk_list[1:], block_size)
    diff_sqr = average([ fsqr(jk - avg) for jk in blocks ])
    fac = math.sqrt(block_size * (len(jk_list) - 1)) / eps
    return fac * fsqrt(diff_sqr)

def jk_avg_err(jk_list, eps = 1, *, block_size = 1):
    return jk_avg(jk_list), jk_err(jk_list, eps, block_size = block_size)

def merge_jk_idx(*jk_idx_list):
    for jk_idx in jk_idx_list:
        assert jk_idx[0] == "avg"
    return [ "avg", ] + [ idx for jk_idx in jk_idx_list for idx in jk_idx[1:] ]

def rejk_list(jk_list, jk_idx_list, all_jk_idx):
    # super jackknife
    assert jk_idx_list[0] == "avg"
    assert all_jk_idx[0] == "avg"
    assert len(jk_idx_list) == len(jk_list)
    assert len(jk_idx_list) <= len(all_jk_idx)
    jk_avg = jk_list[0]
    size_new = len(all_jk_idx)
    i_new = 0
    jk_list_new = []
    for i, idx in enumerate(jk_idx_list):
        while all_jk_idx[i_new] != idx:
            jk_list_new.append(jk_avg)
            i_new += 1
            assert i_new < size_new
        jk_list_new.append(jk_list[i])
        i_new += 1
    while i_new < size_new:
        jk_list_new.append(jk_avg)
        i_new += 1
    assert i_new == size_new
    assert size_new == len(jk_list_new)
    return jk_list_new

# ----------

def mk_jk_blocking_func(block_size = 1, block_size_dict = None):
    if block_size_dict is None:
        block_size_dict = dict()
    def f(idx):
        if isinstance(idx, int):
            traj = idx
            bs = block_size
            return traj // bs
        elif isinstance(idx, tuple) and len(idx) == 2 and isinstance(idx[1], int):
            job_tag, traj = idx
            assert isinstance(traj, int)
            bs = block_size_dict.get(job_tag, block_size)
            assert isinstance(bs, int)
            return (job_tag, traj // bs,)
        else:
            return idx
    return f

def rjk_jk_list(jk_list, jk_idx_list, n_rand_sample, rng_state, jk_blocking_func = None):
    """return rjk_list
    len(rjk_list) == 1 + n_rand_sample
    distribution of rjk_list should be similar as the distribution of avg
    r_{i,j} ~ N(0, 1)
    avg = jk_avg(jk_list)
    len(jk_list) = n + 1
    rjk_list[i] = avg + \\sum_{j=1}^{n} r_{i,j} (jk_list[j] - avg)"""
    assert jk_idx_list[0] == "avg"
    assert isinstance(n_rand_sample, int)
    assert n_rand_sample >= 0
    assert isinstance(rng_state, RngState)
    rs = rng_state
    avg = jk_avg(jk_list)
    n = len(jk_list) - 1
    jk_diff = [ jk_list[j] - avg for j in range(1, n + 1) ]
    rjk_list = [ avg, ]
    if jk_blocking_func is None:
        blocked_jk_idx_list = jk_idx_list
    else:
        blocked_jk_idx_list = [ jk_blocking_func(idx) for idx in jk_idx_list[:] ]
    assert len(blocked_jk_idx_list[1:]) == n
    for i in range(n_rand_sample):
        rsi = rs.split(str(i))
        r = [ rsi.split(str(idx)).g_rand_gen() for idx in blocked_jk_idx_list[1:] ]
        rjk_list.append(avg + sum([ r[j] * jk_diff[j] for j in range(n) ]))
    return rjk_list

def rjk_mk_jk_val(rs_tag, val, err, n_rand_sample, rng_state):
    """return rjk_list
    n = n_rand_sample
    len(rjk_list) == 1 + n
    rjk_list[i] = val + err * r[i] for i in 1..n
    where r[i] ~ N(0, 1)"""
    assert n_rand_sample >= 0
    assert isinstance(rng_state, RngState)
    rjk_list = [ val, ]
    rs = rng_state.split(str(rs_tag))
    for i in range(n_rand_sample):
        r = rs.g_rand_gen()
        rjk_list.append(val + r * err)
    return rjk_list

def rjackknife(data_list, jk_idx_list, n_rand_sample, rng_state, *, eps = 1):
    jk_list = jackknife(data_list, eps)
    return rjk_jk_list(jk_list, jk_idx_list, n_rand_sample, rng_state)

def rjk_avg(rjk_list):
    return jk_avg(rjk_list)

def rjk_err(rjk_list, eps = 1):
    """Return \\frac{1}{eps} \\sqrt{ \\frac{1}{N} \\sum_{i=1}^N (jk[i] - jk_avg)^2 }
    Note: len(jk_list) = N + 1
    Same eps as the eps used in the 'jackknife' function"""
    n = len(rjk_list) - 1
    return jk_err(rjk_list, eps * np.sqrt(n))

def rjk_avg_err(rjk_list, eps = 1):
    return rjk_avg(rjk_list), rjk_err(rjk_list, eps)

# ----------

default_g_jk_kwargs = {}

default_g_jk_kwargs["jk_type"] = "super"  # choices: "rjk", "super"
default_g_jk_kwargs["eps"] = 1

default_g_jk_kwargs["all_jk_idx"] = None # for jk_type = "super"
default_g_jk_kwargs["get_all_jk_idx"] = None # for jk_type = "super"

default_g_jk_kwargs["n_rand_sample"] = 1024 # for jk_type = "rjk"
default_g_jk_kwargs["rng_state"] = RngState("rejk")  # for jk_type = "rjk"
default_g_jk_kwargs["jk_blocking_func"] = None  # for jk_type = "rjk"

@use_kwargs(default_g_jk_kwargs)
@timer
def g_jk(data_list, *, eps, **_kwargs):
    return jackknife(data_list, eps)

@use_kwargs(default_g_jk_kwargs)
def g_jk_blocking_func(idx, *, jk_blocking_func, **_kwargs):
    if jk_blocking_func is None:
        return idx
    else:
        return jk_blocking_func(idx)

@use_kwargs(default_g_jk_kwargs)
def g_mk_jk_val(rs_tag, val, err, n_rand_sample, rng_state, *, jk_type, **_kwargs):
    assert jk_type == "rjk"
    return rjk_mk_jk_val(rs_tag, val, err, n_rand_sample, rng_state)

@use_kwargs(default_g_jk_kwargs)
@timer
def g_rejk(jk_list, jk_idx_list, *,
        jk_type,
        all_jk_idx,
        get_all_jk_idx,
        n_rand_sample,
        rng_state,
        jk_blocking_func,
        **_kwargs,):
    # jk_type in [ "rjk", "super", ]
    if jk_type == "super":
        if jk_blocking_func is None:
            displayln_info("g_rejk: jk_type={jk_type} does not support jk_blocking_func={jk_blocking_func}")
        if all_jk_idx is None:
            assert get_all_jk_idx is not None
            all_jk_idx = get_all_jk_idx()
        return rejk_list(jk_list, jk_idx_list, all_jk_idx)
    elif jk_type == "rjk":
        return rjk_jk_list(jk_list, jk_idx_list, n_rand_sample, rng_state, jk_blocking_func)
    else:
        assert False
    return None

def g_jk_avg(jk_list):
    if isinstance(jk_list, (int, float, complex)):
        return jk_list
    return jk_avg(jk_list)

@use_kwargs(default_g_jk_kwargs)
def g_jk_err(jk_list, *, eps, jk_type, **_kwargs):
    if isinstance(jk_list, (int, float, complex)):
        return 0
    if jk_type == "super":
        return jk_err(jk_list, eps)
    elif jk_type == "rjk":
        return rjk_err(jk_list, eps)
    else:
        assert False
    return None

def g_jk_avg_err(jk_list, **kwargs):
    return g_jk_avg(jk_list), g_jk_err(jk_list, **kwargs)

@use_kwargs(default_g_jk_kwargs)
def g_jk_size(**kwargs):
    jk_type = kwargs["jk_type"]
    all_jk_idx = kwargs["all_jk_idx"]
    get_all_jk_idx = kwargs["get_all_jk_idx"]
    n_rand_sample = kwargs["n_rand_sample"]
    # jk_type in [ "rjk", "super", ]
    if jk_type == "super":
        if all_jk_idx is None:
            assert get_all_jk_idx is not None
            all_jk_idx = get_all_jk_idx()
        return len(all_jk_idx)
    elif jk_type == "rjk":
        return 1 + n_rand_sample
    else:
        assert False
    return None



from libcpp.string cimport string as std_string
from .complex cimport *

cdef extern from "qlat-utils/rng-state.h" namespace "qlat":

    cdef cppclass RngState:
        RngState()
        RngState(const std_string& seed)
        RngState(const RngState& rs0, const std_string& sindex)
        const RngState& operator=(const RngState& rs)
        RngState split(const std_string& sindex)
        RngState newtype(const unsigned long type)

    uint64_t rand_gen(RngState& rs)
    double u_rand_gen(RngState& rs, const double upper, const double lower)
    double g_rand_gen(RngState& rs, const double center, const double sigma)



__all__ = [
        'show_datatable', 'read_datatable',
        'save_datatable', 'load_datatable',
        'azip',
        'plot_save', 'plot_view',
        'gnuplot_png_density',
        'plot_save_display_width',
        'display_img',
        ]

import numpy as np
import os
import sys
import shutil
import tempfile
import subprocess

from qlat_utils.timer import *

def mk_file_dirs(fn):
    path = os.path.dirname(fn)
    if path != "":
        os.makedirs(path, exist_ok=True)

def show_number(x):
    if isinstance(x, complex):
        return f"{x.real} {x.imag}i"
    elif isinstance(x, float):
        return f"{x}"
    elif isinstance(x, float):
        return f"{x}"
    else:
        return str(x)

def show_vector(vec):
    return " ".join([ show_number(x) for x in vec ])

def show_datatable(arr, *, is_return_list_of_string = False):
    lines = [ show_vector(vec) + "\n" for vec in arr ]
    if is_return_list_of_string:
        return lines
    return "".join(lines)

def touch_file(fn, content = "", *, is_directory_exist = False):
    if not is_directory_exist:
        mk_file_dirs(fn)
    with open(fn, "w") as f:
        if isinstance(content, str):
            f.write(content)
        else:
            for s in content:
                f.write(s)

def save_datatable(arr, fn, *, is_directory_exist = False):
    """save_datatable(arr, fn), arr is (numpy) 2-D array, fn is file path name."""
    touch_file(fn,
            show_datatable(arr, is_return_list_of_string = True),
            is_directory_exist = is_directory_exist)

def read_number(s):
    if s[-1] == "i":
        return complex(0, float(s[:-1]))
    else:
        return float(s)

def read_vector(line):
    ss = line.split()
    fs = [ read_number(s) for s in ss if s != "" ]
    n = len(fs)
    xs = []
    i = 0
    while i < n:
        if i + 1 < n and isinstance(fs[i + 1], complex):
            xs.append(fs[i] + fs[i + 1])
            i += 2
        else:
            xs.append(fs[i])
            i += 1
    return xs

def read_datatable(lines):
    # return list of list of numbers (float or complex)
    if isinstance(lines, str):
        return read_datatable(lines.splitlines())
    return [ read_vector(line) for line in lines ]

def load_datatable(fn):
    with open(fn) as f:
        return read_datatable(f)

def azip(vec, *vecs):
    size_list = map(len, vecs)
    size_min = len(vec)
    for size in size_list:
        if size < size_min:
            size_min = size
    return np.array([ v[:size_min] for v in [ vec, ] + list(vecs) ]).transpose()

gnuplot_png_density = 500

def mk_tmp_dir():
    return tempfile.mkdtemp(suffix = ".dir", prefix="pyplot.")

def mk_convert_sh():
    return "\n".join([
        "for i in *.mp ; do",
        "fn=${i%.mp}",
        "rm mpost-job.* 2>&1",
        "TEX=latex mpost -jobname mpost-job $fn.mp",
        "for i in mpost-job.? ; do",
        "echo \"$i\"",
        "mv \"$i\" \"$fn\"-\"${i#mpost-job.}\".eps",
        "done",
        "done",
        "",
        ])

valid_fn_chars = ("abcdefghijklmnopqrstuvwxyz"
        + "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        + "0123456789"
        + " ,.+-_=;:[]?{}"
        )

def check_fn(fn):
    if isinstance(fn, str):
        for c in fn:
            if c not in valid_fn_chars:
                return False
        return True
    return False

def get_plot_name(fn):
    assert fn[-4:] == ".png"
    assert check_fn(fn)
    name = fn[:-4]
    return name

def mk_makefile(fn = None):
    # fn is the target file name, e.g. plot.pdf or plot.png
    if fn is not None:
        name = get_plot_name(fn)
        target = "install"
    else:
        name = "plot-0"
        target = "png"
    return "\n".join([
        f"all: {target}",
        "",
        "gnuplot:",
        "\tgnuplot plotfile",
        "",
        "mpost: gnuplot",
        "\tbash ./convert.sh",
        "",
        "pdf: mpost",
        "\tepstopdf plot-0.eps",
        f"\tmv plot-0.pdf tmp ; mv tmp '{name}.pdf'",
        "",
        "png: pdf",
        f"\tpdftoppm -r {gnuplot_png_density} -png '{name}.pdf' > plot-0.png",
        f"\tmv plot-0.png tmp ; mv tmp '{name}.png'",
        "",
        "install-pdf: pdf",
        f"\tmv '{name}.pdf' ../'{name}.pdf'",
        "",
        "install-png: png",
        f"\tmv '{name}.png' ../'{name}.png'",
        "",
        "install: install-png install-pdf",
        "",
        ])

def mk_plotfile(plot_cmds, plot_lines):
    plot_prefix = [
            "set term mp color latex prologues 3 amstex",
            "set output 'plot.mp'",
            ]
    plot = plot_lines[0] + " \\\n    " + ", \\\n    ".join(plot_lines[1:])
    return "\n".join(plot_prefix + [ "", ] + plot_cmds + [ "", plot, "", ])

def populate_pyplot_folder(
        path,
        *,
        fn = None,
        dict_datatable = None,
        plot_cmds = None,
        plot_lines = None,
        ):
    if dict_datatable is None:
        dict_datatable = {}
    if plot_cmds is None:
        plot_cmds = []
    if plot_lines is not None:
        touch_file(os.path.join(path, "plotfile"),
                mk_plotfile(plot_cmds, plot_lines))
    touch_file(os.path.join(path, "convert.sh"), mk_convert_sh())
    touch_file(os.path.join(path, "Makefile"), mk_makefile(fn))
    for key, dt in dict_datatable.items():
        assert key[-4:] == ".txt"
        assert check_fn(key)
        save_datatable(dt, os.path.join(path, key))

def qremove_all(path):
    return shutil.rmtree(path, ignore_errors = True)

def mk_pyplot_folder(path = None):
    if path is None:
        path = mk_tmp_dir()
    else:
        assert isinstance(path, str)
        assert path.endswith(".pyplot.dir")
        qremove_all(path)
        os.makedirs(path)
    return path

def display_img(fn, *, width = None):
    from IPython.display import HTML, Image, display
    displayln_info(0, f"display_img: fn='{fn}'")
    show_width = ""
    if width is not None:
        show_width = f"width='{width}'"
    try:
      import google.colab
      is_in_colab = True
    except:
      is_in_colab = False
    if not is_in_colab:
        display(HTML(data = f"<img src='{fn}' {show_width} />"))
    else:
        display(Image(filename = fn, width = width))

plot_save_display_width = None

@timer
def plot_save(
        fn = None,
        dts = None,
        cmds = None,
        lines = None,
        *,
        is_run_make = True,
        is_display = False,
        is_verbose = False,
        display_width = None,
        ):
    # fn is full name of the plot or None
    # dts is dict_datatable, e.g. { "table.txt" : [ [ 0, 1, ], [ 1, 2, ], ], }
    # cmds is plot_cmds, e.g. [ "set key rm", "set size 1.0, 1.0 ", ]
    # lines is plot_lines, e.g. [ "plot", "x", ]
    target = fn
    target_fn = None
    path = None
    if target is not None:
        target_fn = os.path.basename(target)
        path = os.path.join(os.path.dirname(target),
                get_plot_name(target_fn) + ".pyplot.dir")
    path = mk_pyplot_folder(path)
    if cmds is None:
        cmds = [
                "set size 0.8, 1.0",
                "set key tm",
                "set xlabel '$x$'",
                "set ylabel '$y$'",
                ]
        displayln_info(0, f"cmds={cmds}")
    if dts is None:
        x = np.arange(31) * (6 / 30) - 3
        y = np.cos(x)
        yerr = 0.1 / (1 + x**2)
        dts = {
                "table.txt" : azip(x, y, yerr),
                }
        displayln_info(f"dts={dts}")
        if lines is None:
            lines = [
                    "plot [-3:3] [-1.5:1.5]",
                    "0 not",
                    "sin(x) w l t '$y = \\sin(x)$'",
                    ]
            if "table.txt" in dts:
                lines.append("'table.txt' w yerrorb t '$y = \\cos(x)$'")
            displayln_info(f"lines={lines}")
    elif lines is None:
        lines = [
                "plot [:] [:]",
                ]
        for key, val in dts.items():
            if len(val[0]) >= 3:
                lines.append(f"'{key}' u 1:2:3 w yerrorb t '{key}'")
            elif len(val[0]) == 2:
                lines.append(f"'{key}' u 1:2 w p t '{key}'")
            else:
                lines.append(f"'{key}' t '{key}'")
        displayln_info(f"lines={lines}")
    if fn is None:
        displayln_info(f"fn={fn}")
        displayln_info(f"is_run_make={is_run_make}")
        displayln_info(f"is_display={is_display}")
    if display_width is None:
        display_width = plot_save_display_width
        displayln_info(0, f"display_width={display_width}")
    populate_pyplot_folder(
            path,
            fn = target_fn,
            dict_datatable = dts,
            plot_cmds = cmds,
            plot_lines = lines,
            )
    if is_run_make:
        @timer
        def qplot_run_make():
            status = subprocess.run([ "make", "-C", path, ], capture_output = True, text = True)
            if is_verbose or status.returncode != 0:
                displayln_info("stdout:")
                displayln_info(status.stdout)
                displayln_info("stderr:")
                displayln_info(status.stderr)
            assert status.returncode == 0
        qplot_run_make()
        if target is None:
            path_img = os.path.join(path, "plot-0.png")
        else:
            path_img = target
        if is_display:
            display_img(path_img, width = display_width)
        else:
            displayln_info(0, f"plot_save: plot created at '{path_img}'.")
        return path_img
    else:
        assert not is_display
        # return directory that contain the sources instead of the png path
        displayln_info(0, f"plot_save: creating data for plot at '{path}'.")
        return path

def plot_view(
        fn = None,
        dts = None,
        cmds = None,
        lines = None,
        *,
        is_verbose = False,
        display_width = None,
        ):
    return plot_save(
            fn = fn, dts = dts, cmds = cmds, lines = lines,
            is_run_make = True,
            is_verbose = is_verbose,
            is_display = True,
            display_width = display_width,
            )



import qlat_utils.c as c

from qlat_utils.c import Coordinate, RngState, WilsonMatrix, SpinMatrix

from qlat_utils.c import \
        ElemTypeColorMatrix, \
        ElemTypeWilsonMatrix, \
        ElemTypeNonRelWilsonMatrix, \
        ElemTypeSpinMatrix, \
        ElemTypeWilsonVector, \
        ElemTypeComplex, \
        ElemTypeComplexF, \
        ElemTypeDouble, \
        ElemTypeFloat, \
        ElemTypeLong, \
        ElemTypeInt64t, \
        ElemTypeInt8t, \
        ElemTypeChar

from qlat_utils.timer import *

from qlat_utils.cache import *

from qlat_utils.qar import *

from qlat_utils.utils import *

from qlat_utils.utils_io import *

from qlat_utils.lat_io import *

from qlat_utils.data import *

from qlat_utils.qplot import *

from qlat_utils.parallel import *

from qlat_utils.get_include_dir import *

del c



# cython: c_string_type=unicode, c_string_encoding=utf8

from . cimport everything as cc
from cpython cimport Py_buffer
from cpython.buffer cimport PyBUF_FORMAT
cimport numpy

import cqlat_utils as c
import numpy as np
import functools

### -------------------------------------------------------------------

cdef class ElemType:

    name = ""

### -------------------------------------------------------------------

# ColorMatrix WilsonMatrix NonRelWilsonMatrix SpinMatrix WilsonVector Complex ComplexF Double Float Long Int64t Int8t Char

cdef class ElemTypeColorMatrix(ElemType):
    name = "ColorMatrix"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'Zd'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Complex)
    @staticmethod
    cdef int ndim():
        return 2
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t](2, 3)
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.ColorMatrix)
cdef class ElemTypeWilsonMatrix(ElemType):
    name = "WilsonMatrix"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'Zd'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Complex)
    @staticmethod
    cdef int ndim():
        return 2
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t](2, 12)
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.WilsonMatrix)
cdef class ElemTypeNonRelWilsonMatrix(ElemType):
    name = "NonRelWilsonMatrix"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'Zd'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Complex)
    @staticmethod
    cdef int ndim():
        return 2
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t](2, 6)
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.NonRelWilsonMatrix)
cdef class ElemTypeSpinMatrix(ElemType):
    name = "SpinMatrix"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'Zd'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Complex)
    @staticmethod
    cdef int ndim():
        return 2
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t](2, 4)
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.SpinMatrix)
cdef class ElemTypeWilsonVector(ElemType):
    name = "WilsonVector"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'Zd'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Complex)
    @staticmethod
    cdef int ndim():
        return 1
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t](1, 12)
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.WilsonVector)
cdef class ElemTypeComplex(ElemType):
    name = "Complex"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'Zd'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Complex)
    @staticmethod
    cdef int ndim():
        return 0
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t]()
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.Complex)
cdef class ElemTypeComplexF(ElemType):
    name = "ComplexF"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'Zf'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.ComplexF)
    @staticmethod
    cdef int ndim():
        return 0
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t]()
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.ComplexF)
cdef class ElemTypeDouble(ElemType):
    name = "Double"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'd'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Double)
    @staticmethod
    cdef int ndim():
        return 0
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t]()
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.Double)
cdef class ElemTypeFloat(ElemType):
    name = "Float"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'f'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Float)
    @staticmethod
    cdef int ndim():
        return 0
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t]()
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.Float)
cdef class ElemTypeLong(ElemType):
    name = "Long"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'q'
        if not sizeof(cc.Long) == 8:
            assert sizeof(cc.Long) == 4
            fmt = 'l'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Long)
    @staticmethod
    cdef int ndim():
        return 0
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t]()
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.Long)
cdef class ElemTypeInt64t(ElemType):
    name = "Int64t"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'q'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Int64t)
    @staticmethod
    cdef int ndim():
        return 0
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t]()
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.Int64t)
cdef class ElemTypeInt8t(ElemType):
    name = "Int8t"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'b'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Int8t)
    @staticmethod
    cdef int ndim():
        return 0
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t]()
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.Int8t)
cdef class ElemTypeChar(ElemType):
    name = "Char"
    @staticmethod
    cdef char* format():
        cdef char* fmt = 'c'
        return fmt
    @staticmethod
    cdef Py_ssize_t itemsize():
        return sizeof(cc.Char)
    @staticmethod
    cdef int ndim():
        return 0
    @staticmethod
    cdef cc.std_vector[Py_ssize_t] shape():
        return cc.std_vector[Py_ssize_t]()
    @staticmethod
    cdef Py_ssize_t size():
        return sizeof(cc.Char)

### -------------------------------------------------------------------

def displayln(level, *args):
    # interpret the first argument as verbose level if it is int
    if isinstance(level, int):
        if level <= verbose_level():
            print(*args, flush=True)
    else:
        print(level, *args, flush=True)

def displayln_info(*args):
    if cc.get_id_node() == 0:
        displayln(*args)

### -------------------------------------------------------------------

cdef class Timer:

    def __cinit__(self, const cc.std_string& fname, cc.bool is_verbose = False):
        self.xx = cc.Timer(fname)

    def start(self):
        self.xx.start(self.is_verbose)

    def stop(self):
        self.xx.stop(self.is_verbose)

### -------------------------------------------------------------------

cdef class TimerNone:

    def __cinit__(self):
        pass

    def start(self):
        pass

    def stop(self):
        pass

### -------------------------------------------------------------------

def get_id_node():
    return cc.get_id_node()

def get_num_node():
    return cc.get_num_node()

def verbose_level(level = None):
    if level is None:
        return cc.verbose_level()
    cdef long* p_ret = &cc.verbose_level()
    p_ret[0] = level
    assert cc.verbose_level() == level
    return level

def get_time():
    return cc.get_time()

def get_start_time(time = None):
    if time is None:
        return cc.get_start_time()
    cdef double* p_ret = &cc.get_start_time()
    p_ret[0] = time
    assert cc.get_start_time() == time
    return time

def get_actual_start_time(time = None):
    if time is None:
        return cc.get_actual_start_time()
    cdef double* p_ret = &cc.get_actual_start_time()
    p_ret[0] = time
    assert cc.get_actual_start_time() == time
    return time

def get_total_time():
    return cc.get_total_time()

def get_actual_total_time():
    return cc.get_actual_total_time()

### -------------------------------------------------------------------

def timer_display(str tag = ""):
    cc.Timer.display(tag)
    cc.flush()

def timer_autodisplay():
    cc.Timer.autodisplay()
    cc.flush()

def timer_display_stack():
    cc.Timer.display_stack()
    cc.flush()

def timer_display_stack_always():
    cc.Timer.display_stack_always()
    cc.flush()

def timer_reset(long max_call_times_for_always_show_info = -1):
    cc.Timer.reset(max_call_times_for_always_show_info)

def timer_fork(long max_call_times_for_always_show_info = -1):
    cc.Timer.fork(max_call_times_for_always_show_info)

def timer_merge():
    cc.Timer.merge()

def flush():
    cc.flush()

### -------------------------------------------------------------------

def timer(func):
    cdef cc.std_string fname = "py:" + func.__name__
    cdef cc.Timer qtimer = cc.Timer(fname)
    @functools.wraps(func)
    def qtimer_func(*args, **kwargs):
        qtimer.start()
        ret = func(*args, **kwargs)
        qtimer.stop()
        return ret
    return qtimer_func

def timer_verbose(func):
    cdef cc.std_string fname = "py:" + func.__name__
    cdef cc.Timer qtimer = cc.Timer(fname)
    @functools.wraps(func)
    def qtimer_func(*args, **kwargs):
        cdef cc.bool is_verbose = True
        qtimer.start(is_verbose)
        ret = func(*args, **kwargs)
        qtimer.stop(is_verbose)
        return ret
    return qtimer_func

def timer_flops(func):
    cdef cc.std_string fname = "py:" + func.__name__
    cdef cc.Timer qtimer = cc.Timer(fname)
    @functools.wraps(func)
    def qtimer_func(*args, **kwargs):
        qtimer.start()
        flops, ret = func(*args, **kwargs)
        qtimer.flops += flops
        qtimer.stop()
        return ret
    return qtimer_func

def timer_verbose_flops(func):
    cdef cc.std_string fname = "py:" + func.__name__
    cdef cc.Timer qtimer = cc.Timer(fname)
    @functools.wraps(func)
    def qtimer_func(*args, **kwargs):
        cdef cc.bool is_verbose = True
        qtimer.start(is_verbose)
        flops, ret = func(*args, **kwargs)
        qtimer.flops += flops
        qtimer.stop(is_verbose)
        return ret
    return qtimer_func

### -------------------------------------------------------------------

cdef class Buffer:

    def __cinit__(self, object obj = None, int ndim = 1, int itemsize = 1):
        self.obj = obj
        self.ndim = ndim
        self.itemsize = itemsize
        self.shape_strides.resize(ndim * 2)

    cdef Py_ssize_t get_len(self):
        cdef int i
        cdef Py_ssize_t ret = 1
        for i in range(self.ndim):
            ret *= self.shape_strides[i]
        return ret

    cdef void set_strides(self):
        cdef Py_ssize_t* shapes = &self.shape_strides[0]
        cdef Py_ssize_t* strides = &self.shape_strides[self.ndim]
        cdef int i
        cdef Py_ssize_t stride = self.itemsize
        for i in range(self.ndim - 1, -1, -1):
            strides[i] = stride
            stride *= shapes[i]

### -------------------------------------------------------------------

cdef class Coordinate:

    def __cinit__(self):
        pass

    def __init__(self, x = None):
        if isinstance(x, Coordinate):
            self.xx = (<Coordinate>x).xx
        elif isinstance(x, list):
            assert len(x) == 4
            self.xx = cc.Coordinate(x[0], x[1], x[2], x[3])
        else:
            assert x is None

    def __imatmul__(self, Coordinate v1):
        self.xx = v1.xx
        return self

    def copy(self, cc.bool is_copying_data = True):
        cdef Coordinate x = Coordinate()
        if is_copying_data:
            x.xx = self.xx
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def list(self):
        return [ self.xx[0], self.xx[1], self.xx[2], self.xx[3], ]

### -------------------------------------------------------------------

cdef class RngState:

    def __cinit__(self):
        self.cdata = <long>&(self.xx)

    def __init__(self, x = None, y = None):
        cdef cc.std_string seed
        if x is None:
            assert y is None
            # make a new rng
            self.xx = cc.RngState()
        elif isinstance(x, RngState):
            if y is None:
                # make a copy of x
                self.xx = (<RngState>x).xx
            else:
                # split into a new rng
                seed = str(y)
                self.xx = cc.RngState((<RngState>x).xx, seed)
        else:
            assert y is None
            # seed a new rng
            seed = str(x)
            self.xx = cc.RngState(seed)

    def __imatmul__(self, RngState v1):
        self.xx = v1.xx
        return self

    def copy(self, cc.bool is_copying_data = True):
        cdef RngState x = RngState()
        if is_copying_data:
            x.xx = self.xx
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def split(self, const cc.std_string& seed):
        cdef RngState x = RngState()
        x.xx = self.xx.split(seed)
        return x

    def rand_gen(self):
        return cc.rand_gen(self.xx)

    def u_rand_gen(self, double upper = 1.0, double lower = 0.0):
        return cc.u_rand_gen(self.xx, upper, lower)

    def g_rand_gen(self, double center = 0.0, double sigma = 1.0):
        return cc.g_rand_gen(self.xx, center, sigma)

    def c_rand_gen(self, Coordinate size):
        # size can be total_site of the lattice
        cdef Coordinate x = Coordinate()
        x.xx = cc.c_rand_gen(self.xx, size.xx)
        return x

    def select(self, list l):
        ri = self.rand_gen() % len(l)
        return l[ri]

    def u_rand_fill(self, arr, double upper = 1.0, double lower = 0.0):
        arr = arr.view(np.float64)
        assert arr.base is not None
        arr = arr.ravel()
        assert arr.base is not None
        return self.u_rand_fill_double(arr, upper, lower)

    def u_rand_fill_double(self, double[:] arr, double upper = 1.0, double lower = 0.0):
        cdef long size = arr.size
        cdef long i
        for i in range(size):
            arr[i] = cc.u_rand_gen(self.xx, upper, lower)

    def g_rand_fill(self, arr, double center = 0.0, double sigma = 1.0):
        arr = arr.view(np.float64)
        assert arr.base is not None
        arr = arr.ravel()
        assert arr.base is not None
        return self.g_rand_fill_double(arr, center, sigma)

    def g_rand_fill_double(self, double[:] arr, double center = 0.0, double sigma = 1.0):
        cdef long size = arr.size
        cdef long i
        for i in range(size):
            arr[i] = cc.g_rand_gen(self.xx, center, sigma)

### -------------------------------------------------------------------

def get_double_sig(x, RngState rs):
    if isinstance(x, LatData):
        arr = rs.np.asarray(x).ravel()
        arr_rand = np.zeros(arr.shape, arr.dtype)
        rs.u_rand_fill(arr_rand, 1.0, -1.0)
        return np.sum(arr, arr_rand)
    else:
        displayln_info("get_double_sig", type(x))
        assert False

### -------------------------------------------------------------------

cdef class LatData:

    def __cinit__(self):
        self.cdata = <long>&(self.xx)

    def __imatmul__(self, LatData v1):
        self.xx = v1.xx
        return self

    def copy(self, cc.bool is_copying_data = True):
        x = LatData()
        if is_copying_data:
            x.xx = self.xx
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef cc.bool is_complex = self.xx.is_complex()
        cdef char* fmt
        cdef int itemsize
        if is_complex:
            fmt = 'Zd'
            itemsize = sizeof(cc.Complex)
        else:
            fmt = 'd'
            itemsize = sizeof(cc.Double)
        cdef Buffer buf = Buffer(self, self.xx.ndim(), itemsize)
        cdef Py_ssize_t* shape = &buf.shape_strides[0]
        cdef Py_ssize_t* strides = &buf.shape_strides[buf.ndim]
        cdef int i
        for i in range(buf.ndim):
            shape[i] = self.xx.info[i].size
        buf.set_strides()
        buffer.buf = <char*>(self.xx.data())
        if flags & PyBUF_FORMAT:
            buffer.format = fmt
        else:
            buffer.format = NULL
        buffer.internal = NULL
        buffer.itemsize = buf.itemsize
        buffer.len = buf.get_len()
        buffer.ndim = buf.ndim
        buffer.obj = buf
        buffer.readonly = 0
        buffer.shape = shape
        buffer.strides = strides
        buffer.suboffsets = NULL

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def save_node(self, const cc.std_string& path):
        self.xx.save(path)

    def load_node(self, const cc.std_string& path):
        self.xx.load(path)

    def save(self, const cc.std_string& path):
        if cc.get_id_node() == 0:
            self.save_node(path)

    def load(self, const cc.std_string& path):
        if cc.get_id_node() == 0:
            self.load_node(path)
        self.bcast()

    def bcast(self):
        if cc.get_num_node() != 1:
            import cqlat as c
            c.bcast_lat_data(self)
        return self

    def glb_sum_in_place(self):
        if cc.get_num_node() != 1:
            import cqlat as c
            c.glb_sum_lat_data(self)
        return self

    def glb_sum(self):
        ld = self.copy()
        return ld.glb_sum_in_place()

    def __str__(self):
        return cc.show(self.xx)

    def show(self):
        return str(self)

    def set_zero(self):
        cc.set_zero(self.xx)

    def qnorm(self):
        return cc.qnorm(self.xx)

    def is_match(self, LatData ld1):
        # ld.info needs to be exactly equal
        return cc.is_matching(self.xx, ld1.xx)

    def is_complex(self):
        return self.xx.is_complex()

    def ndim(self):
        return self.xx.ndim()

    def dim_name(self, int dim):
        assert 0 <= dim
        assert dim < self.xx.info.size()
        return self.xx.info[dim].name

    def dim_size(self, int dim):
        assert 0 <= dim
        assert dim < self.xx.info.size()
        return self.xx.info[dim].size

    def dim_indices(self, int dim):
        assert 0 <= dim
        assert dim < self.xx.info.size()
        return self.xx.info[dim].indices

    def dim_sizes(self):
        cdef int ndim = self.xx.ndim()
        cdef int i
        return [ self.xx.info[i].size for i in range(ndim) ]

    def set_dim_sizes(self, list dim_sizes, *, cc.bool is_complex = True):
        cdef int ndim = len(dim_sizes)
        cdef int ndim_real = ndim
        cdef int i
        if is_complex:
            ndim_real += 1
        self.xx.info.resize(ndim_real)
        for i in range(ndim):
            self.xx.info[i].size = dim_sizes[i]
        if is_complex:
            self.xx.info[ndim] = cc.lat_dim_re_im()
        cc.lat_data_alloc(self.xx)

    def set_dim_name(self, int dim, const cc.std_string& name, list indices = None):
        assert 0 <= dim
        assert dim < self.xx.info.size()
        cdef int size
        cdef int i
        self.xx.info[dim].name = name
        if indices is None:
            self.xx.info[dim].indices.resize(0)
        else:
            indices = [ str(idx).replace("\n", "  ") for idx in indices ]
            size = len(indices)
            self.xx.info[dim].indices.resize(size)
            for i in range(size):
                self.xx.info[dim].indices[i] = indices[i]

    def dim_names(self):
        # by default, return list can be used as the input argument for ld.from_numpy
        cdef int ndim = self.xx.ndim()
        cdef int i
        return [ self.xx.info[i].name for i in range(ndim) ]

    def to_numpy(self):
        return np.asarray(self).copy()

    def from_numpy(self, numpy.ndarray val, list dim_names = None, *, cc.bool is_complex = True):
        # only set LatData shape if it is initially empty
        # otherwise only set data and ignore shape completely
        # dim_names should be a list of names for each dimension
        cdef int ndim = val.ndim
        cdef int dim
        cdef list shape = [ val.shape[dim] for dim in range(ndim) ]
        if self.ndim() == 0:
            self.set_dim_sizes(shape, is_complex = is_complex)
            if dim_names is None:
                dim_names = [ n for n in "ijklmnopqrstuvwxyz" ]
            assert ndim <= len(dim_names)
            for dim in range(ndim):
                name = dim_names[dim]
                self.set_dim_name(dim, name)
        np.asarray(self).ravel()[:] = val.ravel()[:]
        return self

    def to_list(self):
        return np.asarray(self).ravel().tolist()

    def from_list(self, list val, *, cc.bool is_complex = True):
        if self.ndim() == 0:
            self.set_dim_sizes([ len(val), ], is_complex = is_complex)
            self.set_dim_name(0, "i")
        np.asarray(self).ravel()[:] = np.array(val)
        return self

    def __getstate__(self):
        is_complex = self.is_complex()
        ndim = self.ndim()
        dim_sizes = self.dim_sizes()
        assert len(dim_sizes) == ndim
        dim_names = [ self.dim_name(dim) for dim in range(ndim) ]
        dim_indices = [ self.dim_indices(dim) for dim in range(ndim) ]
        data_list = self.to_list()
        return [ is_complex, dim_sizes, dim_names, dim_indices, data_list ]

    def __setstate__(self, state):
        [ is_complex, dim_sizes, dim_names, dim_indices, data_list ] = state
        self.__init__()
        self.set_dim_sizes(dim_sizes, is_complex = is_complex)
        ndim = len(dim_sizes)
        for dim in range(ndim):
            self.set_dim_name(dim, dim_names[dim], dim_indices[dim])
        self.from_list(data_list)

    def info(self, dim = None):
        # by default, return list can be used as the input argument for ld.set_info or mk_lat_data
        if dim is None:
            ndim = self.ndim()
            return [ self.info(i) for i in range(ndim) ]
        else:
            dim_name = self.dim_name(dim)
            dim_size = self.dim_size(dim)
            dim_indices = self.dim_indices(dim)
            return [ dim_name, dim_size, dim_indices, ]

    def set_info(self, list info_list, *, cc.bool is_complex = True):
        # info_list format:
        # [ [ dim_name, dim_size, dim_indices, ], ... ]
        # dim_indices can be optional
        for info in info_list:
            assert len(info) >= 2
        dim_sizes = [ info[1] for info in info_list ]
        self.set_dim_sizes(dim_sizes, is_complex = is_complex)
        ndim = len(dim_sizes)
        for dim in range(ndim):
            info = info_list[dim]
            if len(info) == 2:
                dim_name, dummy_dim_size = info
                self.set_dim_name(dim, dim_name)
            elif len(info) == 3:
                dim_name, dummy_dim_size, dim_indices = info
                self.set_dim_name(dim, dim_name, dim_indices)
            else:
                raise Exception(f"LatData setinfo info_list={info_list}")
        self.set_zero()

    def __iadd__(self, ld1):
        assert isinstance(ld1, LatData)
        c.set_add_lat_data(self, ld1)
        return self

    def __isub__(self, ld1):
        assert isinstance(ld1, LatData)
        c.set_sub_lat_data(self, ld1)
        return self

    def __imul__(self, factor):
        assert isinstance(factor, float)
        c.set_mul_double_lat_data(self, factor)
        return self

    def __add__(self, ld1):
        ld = self.copy()
        ld += ld1
        return ld

    def __sub__(self, ld1):
        ld = self.copy()
        ld -= ld1
        return ld

    def __mul__(self, factor):
        ld = self.copy()
        ld *= factor
        return ld

    __rmul__ = __mul__

    def __neg__(self):
        ld = self.copy()
        ld *= -1
        return ld

    def __pos__(self):
        return self

    def __setitem__(self, idx, val):
        # use list with correct length as val
        # idx should be tuple or list of int
        if isinstance(idx, int):
            idx = [ idx, ]
        if isinstance(val, np.ndarray):
            val = val.flatten()
        c.poke_lat_data(self, idx, list(val))

    def __getitem__(self, idx):
        # return a new list every call
        # idx should be tuple or list of int
        if isinstance(idx, int):
            idx = [ idx, ]
        shape = self.dim_sizes()[len(idx):]
        return np.array(c.peek_lat_data(self, idx)).reshape(shape)

### -------------------------------------------------------------------

def mk_lat_data(list info_list, *, cc.bool is_complex = True):
    # info_list format:
    # [ [ dim_name, dim_size, dim_indices, ], ... ]
    # dim_indices can be optional
    ld = LatData()
    ld.set_info(info_list, is_complex = is_complex)
    return ld

def load_lat_data(const cc.std_string& path):
    ld = LatData()
    ld.load(path)
    return ld

### -------------------------------------------------------------------

cdef class WilsonMatrix:

    def __cinit__(self):
        self.cdata = <long>&(self.xx)

    def __imatmul__(self, WilsonMatrix v1):
        self.xx = v1.xx
        return self

    def copy(self, cc.bool is_copying_data = True):
        cdef WilsonMatrix x = WilsonMatrix()
        if is_copying_data:
            x.xx = self.xx
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def set_zero(self):
        cc.set_zero(self.xx)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef Buffer buf = Buffer(self, ElemTypeWilsonMatrix.ndim(), ElemTypeWilsonMatrix.itemsize())
        cdef char* fmt = ElemTypeWilsonMatrix.format()
        cdef cc.std_vector[Py_ssize_t] vec = ElemTypeWilsonMatrix.shape()
        assert vec.size() == buf.ndim
        cdef Py_ssize_t* shape = &buf.shape_strides[0]
        cdef Py_ssize_t* strides = &buf.shape_strides[buf.ndim]
        cdef int i
        for i in range(buf.ndim):
            shape[i] = vec[i]
        buf.set_strides()
        buffer.buf = <char*>(self.xx.data())
        if flags & PyBUF_FORMAT:
            buffer.format = fmt
        else:
            buffer.format = NULL
        buffer.internal = NULL
        buffer.itemsize = buf.itemsize
        buffer.len = buf.get_len()
        buffer.ndim = buf.ndim
        buffer.obj = buf
        buffer.readonly = 0
        buffer.shape = shape
        buffer.strides = strides
        buffer.suboffsets = NULL

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

    def g5_herm(self):
        self.xx = cc.g5_herm(self.xx)

### -------------------------------------------------------------------

cdef class SpinMatrix:

    def __cinit__(self):
        self.cdata = <long>&(self.xx)

    def __imatmul__(self, SpinMatrix v1):
        self.xx = v1.xx
        return self

    def copy(self, cc.bool is_copying_data = True):
        cdef SpinMatrix x = SpinMatrix()
        if is_copying_data:
            x.xx = self.xx
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def set_zero(self):
        cc.set_zero(self.xx)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef Buffer buf = Buffer(self, ElemTypeSpinMatrix.ndim(), ElemTypeSpinMatrix.itemsize())
        cdef char* fmt = ElemTypeWilsonMatrix.format()
        cdef cc.std_vector[Py_ssize_t] vec = ElemTypeSpinMatrix.shape()
        assert vec.size() == buf.ndim
        cdef Py_ssize_t* shape = &buf.shape_strides[0]
        cdef Py_ssize_t* strides = &buf.shape_strides[buf.ndim]
        cdef int i
        for i in range(buf.ndim):
            shape[i] = vec[i]
        buf.set_strides()
        buffer.buf = <char*>(self.xx.data())
        if flags & PyBUF_FORMAT:
            buffer.format = fmt
        else:
            buffer.format = NULL
        buffer.internal = NULL
        buffer.itemsize = buf.itemsize
        buffer.len = buf.get_len()
        buffer.ndim = buf.ndim
        buffer.obj = buf
        buffer.readonly = 0
        buffer.shape = shape
        buffer.strides = strides
        buffer.suboffsets = NULL

    def __releasebuffer__(self, Py_buffer *buffer):
        pass

### -------------------------------------------------------------------

gamma_matrix_0 = cc.get_gamma_matrix(0)
gamma_matrix_1 = cc.get_gamma_matrix(1)
gamma_matrix_2 = cc.get_gamma_matrix(2)
gamma_matrix_3 = cc.get_gamma_matrix(3)
gamma_matrix_5 = cc.get_gamma_matrix(5)

### -------------------------------------------------------------------

def as_wilson_matrix(x):
    cdef WilsonMatrix wm
    if isinstance(x, WilsonMatrix):
        return x
    elif x == 0:
        wm = WilsonMatrix()
        cc.set_zero(wm.xx)
        return wm

def as_wilson_matrix_g5_herm(x):
    cdef WilsonMatrix wm = WilsonMatrix()
    if isinstance(x, WilsonMatrix):
        wm.xx = cc.g5_herm((<WilsonMatrix>x).xx)
    elif x == 0:
        cc.set_zero(wm.xx)
    return wm

### -------------------------------------------------------------------



from libcpp.vector cimport vector as std_vector

cdef extern from "qlat-utils/handle.h" namespace "qlat":

    cdef cppclass Vector[T]:
        Vector()
        Vector(const T* p, const long n)
        T* data()
        long size()
        long data_size()



from .complex cimport *
from .vector cimport *
from .timer cimport *

cdef extern from "qlat-utils/lat-io.h" namespace "qlat":

    cdef cppclass LatDim:
        std_string name
        long size
        std_vector[std_string] indices
        LatDim()

    LatDim lat_dim_re_im()
    LatDim lat_dim_number(const std_string& name, const long start, const long end)
    LatDim lat_dim_number(const std_string& name, const long start, const long end, const long inc)
    LatDim lat_dim_string(const std_string& name, const std_vector[std_string]& indices)

    ctypedef std_vector[LatDim] LatInfo

    cdef cppclass LatData:
        LatInfo info
        std_vector[double] res
        LatData()
        const LatData& operator=(const LatData& ld)
        void load(const std_string& fn)
        void save(const std_string& fn)
        bool is_complex()
        int ndim()
        double* data()

    bool is_matching(const LatData& ld1, const LatData& ld2)
    long lat_data_size(LatData& ld)
    void lat_data_alloc(LatData& ld)

    Vector[double] get_data(const LatData& x)
    void set_zero(LatData& x)
    std_string show(const LatData& x)
    void clear(LatData& x)
    double qnorm(const LatData& x)



from . import c
from .utils_io import qls
import os

def get_qlat_utils_dir():
    return os.path.dirname(c.__file__)

def get_qlat_utils_include():
    return os.path.join(get_qlat_utils_dir(), 'include')

def get_dir_list():
    return [ get_qlat_utils_dir(), ]

def get_include_list():
    return [ os.path.join(p, 'include') for p in get_dir_list() ]

def get_lib_list():
    return [ os.path.join(p, 'lib') for p in get_dir_list() ]

def get_pxd_list():
    l = []
    for d in get_dir_list():
        fn_list = qls(d)
        for fn in fn_list:
            if fn.endswith(".pxd"):
                l.append(fn)
    return l



from qlat_utils.cache import *
from qlat_utils.rng_state import *
from qlat_utils.utils import *

import sys
import os
import multiprocessing as mp
import gc

pool_function = None

def call_pool_function(*args, **kwargs):
    assert pool_function is not None
    return pool_function(*args, **kwargs)

def process_initialization():
    verbose_level(-1)
    timer_reset(0)
    # gc.unfreeze()
    # clean_cache()
    # gc.collect()
    # gc.freeze()
    # clear_all_caches()

def get_q_num_mp_processes():
    global get_q_num_mp_processes
    s = getenv("q_num_mp_processes", "q_num_threads", "OMP_NUM_THREADS", default = "0")
    v = int(s)
    get_q_num_mp_processes = lambda : v
    return v

def set_q_num_mp_processes(v):
    global get_q_num_mp_processes
    get_q_num_mp_processes = lambda : v

def get_q_verbose_parallel_map():
    global get_q_verbose_parallel_map
    s = getenv("q_verbose_parallel_map", default = "1")
    v = int(s)
    get_q_verbose_parallel_map = lambda : v
    return v

def set_q_verbose_parallel_map(v):
    global get_q_verbose_parallel_map
    get_q_verbose_parallel_map = lambda : v

@timer
def parallel_map(func, iterable,
        _to_be_removed_ = None,
        *,
        n_proc = None,
        chunksize = 1,
        process_initialization = process_initialization,
        verbose = None):
    # iterable = [ i1, i2, ... ]
    # v1 = func(i1)
    # v2 = func(i2)
    # ...
    # return [ v1, v2, ... ]
    #####
    if isinstance(func, int):
        # assuming it is called with parallel_map(n_proc, func, iterable, ...)
        return parallel_map(iterable, _to_be_removed_,
                n_proc = func,
                sum_function = sum_function,
                chunksize = chunksize,
                process_initialization = process_initialization,
                verbose = verbose)
    assert _to_be_removed_ is None
    #####
    if n_proc is None:
        n_proc = get_q_num_mp_processes()
    if verbose is None:
        verbose = get_q_verbose_parallel_map()
    if verbose > 0:
        displayln_info(f"parallel_map(n_proc={n_proc})")
    if n_proc == 0:
        return list(map(func, iterable))
    assert n_proc >= 1
    global pool_function
    assert pool_function is None
    try:
        pool_function = func
        gc.collect()
        gc.freeze()
        with mp.Pool(n_proc, process_initialization, []) as p:
            if verbose > 0:
                p.apply(show_memory_usage)
            res = p.map(call_pool_function, iterable, chunksize = chunksize)
            if verbose > 0:
                p.apply(show_memory_usage)
                p.apply(timer_display)
    finally:
        gc.unfreeze()
        gc.collect()
        pool_function = None
    return res

@timer
def parallel_map_sum(func, iterable,
        _to_be_removed_ = None,
        *,
        n_proc = None,
        sum_function = None,
        sum_start = None,
        chunksize = 1,
        process_initialization = process_initialization,
        verbose = None):
    # iterable = [ i1, i2, ... ]
    # v1 = func(i1)
    # v2 = func(i2)
    # ...
    # return sum_function([ v1, v2, ... ])
    #####
    if isinstance(func, int):
        # assuming it is called with parallel_map_sum(n_proc, func, iterable, ...)
        return parallel_map_sum(iterable, _to_be_removed_,
                n_proc = func,
                sum_function = sum_function,
                sum_start = sum_start,
                chunksize = chunksize,
                process_initialization = process_initialization,
                verbose = verbose)
    assert _to_be_removed_ is None
    #####
    if n_proc is None:
        n_proc = get_q_num_mp_processes()
    if verbose is None:
        verbose = get_q_verbose_parallel_map()
    if verbose > 0:
        displayln_info(f"parallel_map_sum(n_proc={n_proc})")
    if sum_function is None:
        sum_function = sum
    if n_proc == 0:
        return sum_function(map(func, iterable))
    assert n_proc >= 1
    global pool_function
    assert pool_function is None
    try:
        pool_function = func
        gc.collect()
        gc.freeze()
        with mp.Pool(n_proc, process_initialization, []) as p:
            if verbose > 0:
                p.apply(show_memory_usage)
            res = p.imap(call_pool_function, iterable, chunksize = chunksize)
            if sum_start is None:
                ret = sum_function(res)
            else:
                ret = sum_function(res, sum_start)
            if verbose > 0:
                p.apply(show_memory_usage)
                p.apply(timer_display)
    finally:
        gc.unfreeze()
        gc.collect()
        pool_function = None
    return ret

def sum_list(res, start = None):
    # res = [ [ va1, vb1, ... ], [ va2, vb2, ... ], ... ]
    # return [ sum([va1, va2, ...]), sum([vb1, vb2, ...]), ... ]
    if start is not None:
        ret = list(start)
    else:
        ret = None
    for r in res:
        if ret is None:
            ret = list(r)
            continue
        for i, v in enumerate(r):
            ret[i] += v
    return ret



import qlat_utils.c as c

import os
import pickle

from qlat_utils.c import qremove
from qlat_utils.c import qremove_all
from qlat_utils.c import qmkdir, qmkdir_info
from qlat_utils.c import does_file_exist
from qlat_utils.c import is_directory
from qlat_utils.c import is_regular_file
from qlat_utils.c import qrename, qrename_info
from qlat_utils.c import qls
from qlat_utils.c import qls_all

from qlat_utils.qar import *

@timer
def qmkdirs(path):
    os.makedirs(path, exist_ok=True)

@timer
def qmkdirs_info(path):
    if get_id_node() == 0:
        displayln(f"qmkdirs_info: '{path}'.")
        qmkdirs(path)

@timer
def mk_dirs(path):
    os.makedirs(path, exist_ok=True)

@timer
def mk_dirs_info(path):
    if get_id_node() == 0:
        displayln(f"mk_dirs_info: '{path}'.")
        mk_dirs(path)

@timer
def mk_file_dirs(fn):
    path = os.path.dirname(fn)
    if path != "":
        os.makedirs(path, exist_ok=True)

@timer
def mk_file_dirs_info(path):
    if get_id_node() == 0:
        displayln(f"mk_file_dirs_info: '{path}'.")
        mk_file_dirs(path)

def qtouch(path, content = None):
    # mk_file_dirs(path)
    if content is None:
        return c.qtouch(path)
    else:
        return c.qtouch(path, content)

def qtouch_info(path, content = None):
    # mk_file_dirs_info(path)
    if content is None:
        return c.qtouch_info(path)
    else:
        return c.qtouch_info(path, content)

def qappend(path, content = None):
    # mk_file_dirs(path)
    if content is None:
        return c.qappend(path)
    else:
        return c.qappend(path, content)

def qappend_info(path, content = None):
    # mk_file_dirs_info(path)
    if content is None:
        return c.qappend_info(path)
    else:
        return c.qappend_info(path, content)

@timer
def compute_crc32(path):
    return c.compute_crc32(path)

def qload_datatable(path, is_par = False):
    return c.qload_datatable(path, is_par)

@timer
def check_all_files_crc32_info(path):
    return c.check_all_files_crc32_info(path)

@timer
def save_pickle_obj(obj, path):
    # only save from node 0
    # mk_file_dirs_info(path)
    if get_id_node() == 0:
        qtouch(path, pickle.dumps(obj))

@timer
def load_pickle_obj(path, default_value = None):
    # all the nodes read the same data
    if does_file_exist_qar_sync_node(path):
        obj = pickle.loads(qcat_bytes_sync_node(path))
        return obj
    else:
        return default_value

@timer
def pickle_cache_call(func, path):
    if not does_file_exist_qar_sync_node(path):
        obj = func()
        save_pickle_obj(obj, path)
    else:
        obj = load_pickle_obj(path)
    return obj

def qremove_info(path):
    if get_num_node() != 1:
        import qlat.c as c
        return c.qremove_info(path)
    return qremove(path)

def qremove_all_info(path):
    if get_num_node() != 1:
        import qlat.c as c
        return c.qremove_all_info(path)
    return qremove_all(path)

def qmkdir_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.qmkdir_sync_node(path)
    return qmkdir(path)

def does_file_exist_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.does_file_exist_sync_node(path)
    return does_file_exist(path)

def is_directory_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.is_directory_sync_node(path)
    return is_directory(path)

def is_regular_file_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.is_regular_file_sync_node(path)
    return is_regular_file(path)

def qls_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.qls_sync_node(path)
    return qls(path)

def qls_all_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.qls_all_sync_node(path)
    return qls_all(path)

def qload_datatable_sync_node(path, is_par = False):
    if get_num_node() != 1:
        import cqlat as c
        return c.qload_datatable_sync_node(path, is_par)
    return qload_datatable(path, is_par)



from qlat_utils.lat_io cimport *

cdef extern from "qlat-utils/matrix.h" namespace "qlat":

    cdef cppclass ColorMatrix:
        ColorMatrix()
        const ColorMatrix& operator=(const ColorMatrix& m)
        Complex* data()

    cdef cppclass SpinMatrix:
        SpinMatrix()
        const SpinMatrix& operator=(const SpinMatrix& m)
        Complex* data()

    const SpinMatrix operator*(const Complex& a, const SpinMatrix& m)
    const SpinMatrix operator*(const SpinMatrix& m, const Complex& a)
    const SpinMatrix operator*(const SpinMatrix& m1, const SpinMatrix& m2)

    const SpinMatrix& get_gamma_matrix(const int mu)

    cdef cppclass WilsonMatrix:
        WilsonMatrix()
        const WilsonMatrix& operator=(const WilsonMatrix& m)
        Complex* data()

    const WilsonMatrix operator*(const Complex& a, const WilsonMatrix& m)
    const WilsonMatrix operator*(const WilsonMatrix& m, const Complex& a)
    const WilsonMatrix operator*(const WilsonMatrix& m1, const WilsonMatrix& m2)
    const WilsonMatrix operator*(const SpinMatrix& m1, const WilsonMatrix& m2)
    const WilsonMatrix operator*(const WilsonMatrix& m1, const SpinMatrix& m2)

    cdef cppclass NonRelWilsonMatrix:
        NonRelWilsonMatrix()
        const NonRelWilsonMatrix& operator=(const NonRelWilsonMatrix& m)
        Complex* data()

    cdef cppclass WilsonVector:
        WilsonVector()
        const WilsonVector& operator=(const WilsonVector& m)
        Complex* data()

cdef extern from "qlat-utils/lib/mat.h" namespace "qlat":

    void set_zero(ColorMatrix& x)
    void set_zero(SpinMatrix& x)
    void set_zero(WilsonMatrix& x)
    void set_zero(NonRelWilsonMatrix& x)
    void set_zero(WilsonVector& x)

    Vector[Complex] get_data(const ColorMatrix& x)
    Vector[Complex] get_data(const SpinMatrix& x)
    Vector[Complex] get_data(const WilsonMatrix& x)
    Vector[Complex] get_data(const NonRelWilsonMatrix& x)
    Vector[Complex] get_data(const WilsonVector& x)

    Complex mat_tr(const SpinMatrix& m)
    Complex mat_tr(const WilsonMatrix& m)
    Complex mat_tr(const WilsonMatrix& m1, const WilsonMatrix& m2)
    Complex mat_tr(const WilsonMatrix& m1, const SpinMatrix& m2)
    Complex mat_tr(const SpinMatrix& m1, const WilsonMatrix& m2)
    Complex mat_tr(const SpinMatrix& m1, const SpinMatrix& m2)

    WilsonMatrix g5_herm(const WilsonMatrix& m)



import qlat_utils.c as c

from qlat_utils.c import get_qar_multi_vol_max_size
from qlat_utils.c import does_regular_file_exist_qar
from qlat_utils.c import does_file_exist_qar
from qlat_utils.c import qcat
from qlat_utils.c import qcat_bytes

from qlat_utils.timer import *

@timer
def qar_create(path_qar, path_folder, *, is_remove_folder_after = False):
    return c.qar_create(path_qar, path_folder, is_remove_folder_after)

@timer
def qar_extract(path_qar, path_folder, *, is_remove_qar_after = False):
    return c.qar_extract(path_qar, path_folder, is_remove_qar_after)

@timer
def qcopy_file(path_src, path_dst):
    return c.qcopy_file(path_src, path_dst)

@timer
def list_qar(path_qar):
    return c.list_qar(path_qar)

def does_regular_file_exist_qar_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.does_regular_file_exist_qar_sync_node(path)
    return does_regular_file_exist_qar(path)

def does_file_exist_qar_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.does_file_exist_qar_sync_node(path)
    return does_file_exist_qar(path)

def qcat_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.qcat_sync_node(path)
    return qcat(path)

def qcat_bytes_sync_node(path):
    if get_num_node() != 1:
        import cqlat as c
        return c.qcat_bytes_sync_node(path)
    return qcat_bytes(path)

@timer
def qar_create_info(path_qar, path_folder, *, is_remove_folder_after = False):
    if get_num_node() != 1:
        import cqlat as c
        return c.qar_create_info(path_qar, path_folder, is_remove_folder_after)
    return qar_create(path_qar, path_folder, is_remove_folder_after = is_remove_folder_after)

@timer
def qar_extract_info(path_qar, path_folder, *, is_remove_qar_after = False):
    if get_num_node() != 1:
        import cqlat as c
        return c.qar_extract_info(path_qar, path_folder, is_remove_qar_after)
    return qar_extract(path_qar, path_folder, is_remove_qar_after = is_remove_qar_after)

@timer
def qcopy_file_info(path_src, path_dst):
    if get_num_node() != 1:
        import cqlat as c
        return c.qcopy_file_info(path_src, path_dst)
    return qcopy_file(path_src, path_dst)



from . cimport everything as cqlat_utils

cdef class ElemType:

    pass

cdef class ElemTypeColorMatrix(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeWilsonMatrix(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeNonRelWilsonMatrix(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeSpinMatrix(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeWilsonVector(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeComplex(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeComplexF(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeDouble(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeFloat(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeLong(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeInt64t(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeInt8t(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()
cdef class ElemTypeChar(ElemType):
    @staticmethod
    cdef char* format()
    @staticmethod
    cdef Py_ssize_t itemsize()
    @staticmethod
    cdef int ndim()
    @staticmethod
    cdef cqlat_utils.std_vector[Py_ssize_t] shape()
    @staticmethod
    cdef Py_ssize_t size()

### -------------------------------------------------------------------

cdef class Timer:

    cdef cqlat_utils.Timer xx

    cdef cqlat_utils.bool is_verbose

cdef class TimerNone:

    pass

cdef class Buffer:

    cdef object obj

    cdef int ndim

    cdef Py_ssize_t itemsize

    cdef cqlat_utils.std_vector[Py_ssize_t] shape_strides # shape.size() == 2 * ndim

    cdef Py_ssize_t get_len(self)

    cdef void set_strides(self)

cdef class Coordinate:

    cdef cqlat_utils.Coordinate xx

cdef class RngState:

    cdef cqlat_utils.RngState xx

    cdef readonly long cdata

cdef class LatData:

    cdef cqlat_utils.LatData xx

    cdef readonly long cdata

cdef class WilsonMatrix:

    cdef cqlat_utils.WilsonMatrix xx

    cdef readonly long cdata

cdef class SpinMatrix:

    cdef cqlat_utils.SpinMatrix xx

    cdef readonly long cdata

cdef cqlat_utils.SpinMatrix gamma_matrix_0
cdef cqlat_utils.SpinMatrix gamma_matrix_1
cdef cqlat_utils.SpinMatrix gamma_matrix_2
cdef cqlat_utils.SpinMatrix gamma_matrix_3
cdef cqlat_utils.SpinMatrix gamma_matrix_5



cimport libcpp.complex
from libc.stdint cimport uint64_t
from libc.stdint cimport int64_t
from libc.stdint cimport uint8_t
from libc.stdint cimport int8_t

ctypedef double complex Complex
# ctypedef libcpp.complex.complex[double] Complex

ctypedef float complex ComplexF
# ctypedef libcpp.complex.complex[float] Complex

ctypedef double Double

ctypedef float Float

ctypedef long Long

ctypedef int64_t Int64t

ctypedef int8_t Int8t

ctypedef char Char
